/* Public domain code by Sam Trenholme 2000 */
/* Routines in the string library not dependent on the underlying OS */
/* Nor dependent on the encoding of the data in the string--all of these
   work on the underlying binary data in the string */

/* Headers for the string routines */
#include "JessStr.h"

/* js_octets: Number of allowed octets in a given string 
   input: Pointer to the string in question
   output: Number of allowed octets */
int js_octets(JessicaString *js) {
    return js->max_count * js->unit_size;
    }

/* js_init: Create a new string object
   input: The maximum unit count and unit size of the new string object
   output: A pointer to the new string object */

JessicaString *js_create(int max_count, int unit_size) {
    JessicaString *new;

    if((new = js_alloc(1,sizeof(JessicaString))) == (void *)-1) 
        return (JessicaString *)-1;

    /* Initial values for the new string object */
    new->unit_size = unit_size;
    new->unit_count = 0;
    new->max_count = max_count;
    new->is_good = JS_SANE_NUMBER; 
    new->encoding = JS_BINARY;
    /* Allocate memory for character string, return on error */
    if((new->string = js_alloc(max_count,unit_size)) == (void *)-1) {
        js_dealloc(new);
        return (JessicaString *)-1;
        }

    return new;
    }

/* js_set_encode: Change the encoding of a JessicaString object 
   input: Pointer to JessicaString object, desired encoding
   output: -1 on error, otherwise 0
*/
int js_set_encode(JessicaString *js, int encoding) {
    /* Make sure we are sane */
    if(js_has_sanity(js) == -1)
        return -1;
   
    /* Return error if encoding value is out of bounds */
    if(encoding < JS_MIN_ENCODE || encoding > JS_MAX_ENCODE)
        return -1;

    js->encoding = encoding;

    return 0;
    }

/* js_get_encode: Get the encoding of a JessicaString object 
   input: Pointer to JessicaString object
   output: -1 on error, otherwise encoding
*/

int js_get_encode(JessicaString *js) {
    /* Make sure we are sane */
    if(js_has_sanity(js) == -1)
        return -1;
   
    /* Return error if encoding value is out of bounds */
    if(js->encoding < JS_MIN_ENCODE || js->encoding > JS_MAX_ENCODE)
        return -1;

    return js->encoding;

    return 0;
    }

/* js_set_chsize: Change the size of a character in a JessicaString object 
   input: Pointer to JessicaString object, desired character size
   output: -1 on error, otherwise 0
*/
int js_set_chsize(JessicaString *js, int size) {
    /* Make sure we are sane */
    if(js_has_sanity(js) == -1)
        return -1;
   
    /* Return error if size value is out of bounds */
    if(size < 1)
        return -1;

    /* You are only allowed to change the unit size if the string is a
       0-length string */
    if(js->unit_count != 0)
        return -1;

    js->unit_size = size;

    return 0;
    }

/* js_destroy: Destroy a new string object 
   input: Pointer to the string object to destroy
   output: -1 on error, 1 on success*/

int js_destroy(JessicaString *object) {
    if(js_has_sanity(object) != -1)
        return js_destroy_force(object);
    else
        return -1;
    }

/* js_destroy_force: Destroy a new string object (forcibly, may segfault)
   input: Pointer to the string object to destroy
   output: 1 on success */

int js_destroy_force(JessicaString *object) {
    /* Deallocate the character string in the string object */
    js_dealloc(object->string);
    /* Deallocate the object itself */
    js_dealloc(object);
    /* Bye bye */
    return 1;
    }

/* js_has_sanity: Make sure a JessicaString object is sane
   input: pointer to JessicaString object
   output: -1 if insane, otherwise 1
*/

int js_has_sanity(JessicaString *object) {
    /* Is the sane value sane? */
    if(object->is_good != JS_SANE_NUMBER)
        return -1;

    /* Is the length not greater than the max length? */
    if(object->unit_count > object->max_count)
        return -1;

    return 1;
    }

/* js_str2js: Convert a 'normal' C-lang string to a JessicaString object
   input: pointer to js object, string to convert, units to convert, size
          of each unit in octets
   output: -1 on error */

int js_str2js(JessicaString *js, char *string, int count, int size) {
    int counter = 0;
    /* int max = (count > js->max_count ? js->max_count : count); */
    int max = count;

    /* If we are out of bounds, return error */
    if(count > js->max_count) 
        return -1;

    /* Sanity checks */
    if(js_has_sanity(js) == -1)
        return -1;
    if(size != js->unit_size)
        return -1;

    while(counter < (max * size)) {
        *(js->string + counter) = *(string + counter);
	counter++;
        }

    /* Reset the count of characters in the string */
    js->unit_count = count;

    return 1;
    }

/* js_issame: Determine if two JessicaString objects are identical
   input: Pointers to the two string objects
   output: 1 if they are the same, 0 otherwise, -1 on error */

int js_issame(JessicaString *js1, JessicaString *js2) {
    int counter = 0;
    int max;

    /* Sanity checks */
    if(js_has_sanity(js1) == -1)
        return -1;
    if(js_has_sanity(js2) == -1)
        return -1;

    /* They are not the same if they have different sizes for a character */
    if(js1->unit_size != js2->unit_size)
        return 0;
    /* Nor are they the same if they have different lengths */
    if(js1->unit_count != js2->unit_count)
        return 0;

    max = js1->unit_count;

    /* If any characters in the actual string differer, they are different */
    while(counter < (max * js1->unit_size)) {
        if(*(js1->string + counter) != *(js2->string + counter)) 
            return 0; 
        counter++;
	}
   
    /* Otherwise, they are identical */
    return 1;
    }

/* js_fgrep: Determine if the contents of one JessicaString object is
             embedded in another JessicaString object
   input: Object with string to look for, string object to look in
   output: number of units from beginning of string grepped string is
           at if found, 0 otherwise, -1 on error */

int js_fgrep(JessicaString *exp, JessicaString *js) {
    return js_fgrep_offset(exp,js,0);
    }

/* js_fgrep_offset: Determine if the contents of one JessicaString object is
             embedded in another JessicaString object, but we only check 
             starting at offset (0 to look at entire string)
   input: Object with string to look for, string object to look in, offset of
          where we first do the match (in units)
   output: number of units from beginning of string grepped string is
           at if found, -2 otherwise, -1 on error */

int js_fgrep_offset(JessicaString *exp, JessicaString *js, int offset) {
    int counter = 0;
    int place = 0;
    int match = -2;
    int max;

    /* Sanity checks */
    if(js_has_sanity(exp) == -1)
        return -1;
    if(js_has_sanity(js) == -1)
        return -1;

    /* Make sure offset has an OK value; error if it is longer then string */
    if(offset > js->unit_count)
        return -1;

    /* If offset is less than zero, mark it from end of string */
    if(offset < 0)
        offset = js->unit_count + offset;

    /* If offset is still less than zero, error condition */
    if(offset < 0)
        return -1;

    /* They are not the same if they have different sizes for a character */
    if(exp->unit_size != js->unit_size)
        return -2;

    /* They do not match if the grep expression is longer than the string
       to grep from */
    if(exp->unit_count > js->unit_count)
        return -2;

    /* In the case of an empty grep expression, they match (at the top) */
    if(exp->unit_count <= 0)
        return 0;

    /* Max here is the last letter a matching expression can start at
       in the string we are looking for the expression in */
    max = js->unit_count - exp->unit_count;

    match = -2;

    /* Here is the actual grepping core; we match the leftmost instance */
    while(counter <= (max * js->unit_size) && match == -2) {
        /* If we match at the beginning of a letter */
        if(*(exp->string) == *(js->string + counter) &&
          counter % js->unit_size == 0 && counter >= offset * js->unit_size) {
	    /* We make sure we match for each letter in the expression */
	    match = counter / js->unit_size;
	    place = 0;
	    while(place < (exp->unit_count * exp->unit_size) &&
	      place + counter < (js->unit_count * js->unit_size)) {
	        if(*(exp->string + place) != 
		  *(js->string + place + counter))
		    match = -2;
                place++;
		}
            }
        counter++;
	}
   
    /* Return whether they matched or not */
    return match;

    }

/* js_match: Match one of a list of characters
   input: list of characters to match against, string to look in,
   output: -2 if no match, -1 error, otherwise characters form beginning
           match is at */

int js_match(JessicaString *exp, JessicaString *js) {
    return js_match_offset(exp,js,0);
    }

/* js_match_offset: Match one of a list of characters
   input: list of characters to match against, string to look in,
          offset (first character to match from)
   output: -2 if no match, -1 error, otherwise characters from beginning
           match is at */

int js_match_offset(JessicaString *exp, JessicaString *js, int offset) {
    int counter = 0;
    int place = 0;
    int pexp = 0;
    int match = -2;

    /* Sanity checks */
    if(js_has_sanity(exp) == -1)
        return -1;
    if(js_has_sanity(js) == -1)
        return -1;

    /* If offset is less than zero, mark it from end of string */
    if(offset < 0)
        offset = js->unit_count + offset;

    /* If offset is still less than zero, error condition */
    if(offset < 0)
        return -1;

    /* They are not the same if they have different sizes for a character */
    if(exp->unit_size != js->unit_size)
        return -2;

    /* In the case of an empty grep expression, they do not match */
    if(exp->unit_count <= 0)
        return -1;

    /* Here is the actual grepping core; we match the leftmost instance */
    while(counter <= (js->unit_count * js->unit_size) && match == -2) {
        /* Point to first letter in expression */
        pexp = 0;
	/* If we are where a match is allowed (after offset) and the 
	   pexp guy (Pointer to where we are looking in the expression
	   right now) is in the expression string */
        while(counter >= offset && pexp < (exp->unit_count * exp->unit_size) 
              && match == -2) {
	    /* If the octet we are looking at in the expression matches
	       of the octet we are looking at in the string; and the octet
	       we are looking at is the first character of a possible
	       multioctet set; and we are in bounds... */
	    if(*(exp->string + pexp) == *(js->string + counter) &&
	      pexp % exp->unit_size == 0 && pexp < js_octets(exp)) {
	        /* ... then we tenatively match */
	        match = counter / exp->unit_size;
	        place = 1;
		/* For multioctet characters, we have to make sure
		   we match for all the octets in the character */
                while(place < exp->unit_size) {
		    /* If we do not match for any of the octets...
		       (and are in bounds) */
		    if(*(exp->string + pexp + place) != 
		      *(js->string + counter + place) && 
		      pexp + place < js_octets(exp) &&
		      counter + place < js_octets(js))
		        /* ... Disregard the match */
		        match = -2;
                    place++;
		    }
                }
            /* Continue looking at all the characters in the expression */
	    pexp++; 
            } 
        /* And continue looking at all the characters in the string we are
	   looking at */
        counter++;
	}
   
    /* Return whether they matched or not */
    return match;

    }

/* js_insert: insert one string inside another string
   input: A pointer of the string obj to insert, the target to insert in,
          and the place we insert the string (in units, unit 0 is beginning
          of string, negative number is units from the end of the string;
           -1 is one from end of string; -2 two from end of string and so on)
   output: -1 on error, 0 on success
*/

int js_insert(JessicaString *toinsert, JessicaString *target, int where) {
    int place = 0;
    int counter;
    char *temp_str; /* temporary string for properly performing insertion */

    /* Check to make sure things are sane */
    if(js_has_sanity(toinsert) == -1)
        return -1;
    if(js_has_sanity(target) == -1)
        return -1;
    /* Both the source and dest must have same unit size */
    if(toinsert->unit_size != target->unit_size)
        return -1;
    /* Where must be before the end of the string */
    if(where > target->unit_count)
        return -1;

    /* If where is precisely at the end, do the more efficient append */
    if(where == target->unit_count)
        return js_append(toinsert,target);

    /* So we don't have to deal with it, bail if the resulting string is 
       longer than the maximum allowed target string */
    if(toinsert->unit_count + target->unit_count > target->max_count)
        return -1;
    
    /* Allocate temporary string; Return on allocation error */
    if((temp_str = js_alloc(target->unit_size,target->unit_count)) == 
       (void *)-1)
        return -1;

    /* Copy over the target string to temp */
    while(place < (target->unit_size * target->unit_count)) {
        *(temp_str + place) = *(target->string + place);
	place++;
	}

    /* If where is less than zero, offset it from the end of the string */
    if(where < 0) 
        where = target->unit_count + where;

    /* If where is still less than zero, return error */
    if(where < 0) {
        js_dealloc(temp_str);
        return -1;
        }

    /* Change the actual target string */
    place = 0;
    counter = 0;
    while(place < (target->unit_size * target->unit_count)) {
        /* If we are where to insert toinsert string, do so */
        if(place == where * target->unit_size) 
	    while(counter < toinsert->unit_size * toinsert->unit_count) {
                /* If we are in the bounds of the string... */
	        if(place + counter < (target->unit_size * target->max_count))
                    /* ...then add an octet of the string to insert */
		    *(target->string + place + counter) = 
		      *(toinsert->string + counter);
                counter++;
		}
        /* If we are in the bounds of the string... */
        if(place + counter < (target->unit_size * target->max_count))
            /* Then add an octet of the original string */
            *(target->string + place + counter) = *(temp_str + place);
        place++;
        }

    /* Increase the size of the string */
    target->unit_count += toinsert->unit_count;

    /* Free that string */
    js_dealloc(temp_str);

    return 0;
    }


/* js_append: add one string to the end of another string
   input: A pointer of the string obj to append, the target to append to,
   output: -1 on error, 0 on success
*/

int js_append(JessicaString *toappend, JessicaString *target) {
    int counter = 0;

    /* Check to make sure things are sane */
    if(js_has_sanity(toappend) == -1)
        return -1;
    if(js_has_sanity(target) == -1)
        return -1;
    /* Both the source and dest must have same unit size */
    if(toappend->unit_size != target->unit_size)
        return -1;

    /* So we don't have to deal with it, bail if the resulting string is 
       longer than the maximum allowed target string */
    if(toappend->unit_count + target->unit_count > target->max_count)
        return -1;
    
    /* Append the toappend string to the actual target string */
    while(counter < (toappend->unit_size * toappend->unit_count)) {
        *(target->string + target->unit_size * target->unit_count + counter) = 
          *(toappend->string + counter);
        counter++;
        }

    /* Increase the size of the string */
    target->unit_count += toappend->unit_count;

    return 0;
    }

/* js_substr: Copy a substring of the first string in to the second string
   input: The source string, the destination string, the first character to
          copy from, the length of the string to copy (in units, possibly
          multi-octet)
   output: 1 on success, -1 on error */
int js_substr(JessicaString *source, JessicaString *dest, int start, int count)
    {
    int counter;
    
    /* Sanity Checks */
    if(js_has_sanity(source) == -1)
        return -1;
    if(js_has_sanity(dest) == -1)
        return -1;
    if(source->unit_size != dest->unit_size)
        return -1;

    /* Make sure the passed parameters are in bounds */
    if(count < 0)
        return -1;
    if(start >= source->unit_count || start >= source->max_count)
        return -1;
    if(start + count >= source->unit_count)
        return -1;
    if(start + count >= source->max_count)
        return -1;
    if(count >= dest->max_count)
        return -1;

    /* Do the actual copying */
    counter = 0;
    while(counter < count * source->unit_size && counter < js_octets(dest) &&
          counter + start * source->unit_size < js_octets(source)) {
        *(dest->string + counter) = 
	  *(source->string + counter + start * source->unit_size);
        counter++;
	}

    dest->unit_count = count;

    }

